from pymongo import MongoClient
from datetime import datetime
from pathlib import Path


class Db:
    def __init__(self, url: str, db_name: str, batch_collection_name: str, file_collection_name: str = "file_infos"):
        self.url = url
        self.db_name = db_name
        self.batch_collection_name = batch_collection_name
        self.file_collection_name = file_collection_name

    def _connect(self):
        try:
            client = MongoClient(self.url)
            return client
        except Exception as e:
            print(f"Error connecting to database: {e}")
            return None

    def _close(self, client: MongoClient):
        if client:
            client.close()
        
    def flag_payloads(self, file_path: Path, flag: str, clean: bool = False):
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.batch_collection_name]

                if clean:
                    filter_query = {"file_name": file_path.name, "status": "WRITTEN"}
                else:
                    filter_query = {"file_name": file_path.name}

                result = collection.update_many(
                    filter_query,
                    {"$set": {"status": flag, "updated_at": datetime.now()}}
                )
                if result.matched_count:
                    print(f"All payloads for file {file_path.name} flagged as {flag}.")
                else:
                    print(f"No payloads found for file {file_path.name}.")
            except Exception as e:
                print(f"Error flagging payloads: {e}")
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")
        
    def save_file(self, **kwargs):
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.file_collection_name]
                file_data = dict(kwargs)
                file_data["created_at"] = datetime.now()
                collection.insert_one(file_data)
                print("File data saved to database.")
            except Exception as e:
                print(f"Error saving file data to database: {e}")
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")
        
    def save_payload(self, **kwargs)->bool:
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.batch_collection_name]
                payload = dict(kwargs)
                payload["status"] = "PENDING"
                payload["created_at"] = datetime.now()
                collection.insert_one(payload)
                print("Payload saved to database.")
                return True
            except Exception as e:
                print(f"Error saving payload to database: {e}")
                return False
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")
            return False

    def get_file(self, file_path: Path):
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.file_collection_name]
                file_info = collection.find_one({"file_name": file_path.name})
                return file_info
            except Exception as e:
                print(f"Error retrieving file info: {e}")
                return None
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")
            return None

    def update_file(self, file_path: Path, **kwargs):
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.file_collection_name]
                update_fields = {k: v for k, v in kwargs.items()}
                update_fields["updated_at"] = datetime.now()
                result = collection.update_one(
                    {"file_path": file_path.name},
                    {"$set": update_fields}
                )
                if result.matched_count:
                    print(f"File {file_path.name} updated with fields {list(kwargs.keys())}.")
                else:
                    print(f"No file info found with file_name {file_path.name}.")
            except Exception as e:
                print(f"Error updating file info: {e}")
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")
        
    def get_payload(self, custom_id: str):
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.batch_collection_name]
                payload = collection.find_one({"custom_id": custom_id})
                return payload
            except Exception as e:
                print(f"Error retrieving payload by custom_id: {e}")
                return None
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")
            return None
        
    def get_payloads(self, status:str)->list:
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.batch_collection_name]
                payloads = list(collection.find({"status": status}))
                return payloads
            except Exception as e:
                print(f"Error retrieving payloads from database: {e}")
                return []
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")
            return []
        
    def update_payload(self, custom_id: str, **kwargs):
        client = self._connect()
        if client:
            try:
                db = client[self.db_name]
                collection = db[self.batch_collection_name]
                update_fields = {k: v for k, v in kwargs.items()}
                update_fields["updated_at"] = datetime.now()
                result = collection.update_one(
                    {"custom_id": custom_id},
                    {"$set": update_fields}
                )
                if result.matched_count:
                    print(f"Payload {custom_id} updated with fields {list(kwargs.keys())}.")
                else:
                    print(f"No payload found with custom_id {custom_id}.")
            except Exception as e:
                print(f"Error updating payload: {e}")
            finally:
                self._close(client)
        else:
            print("Failed to connect to the database.")


from .db import Db

class Line:

    def __init__(
        self,
        db: Db,
        custom_id: str,
        user_prompt: str,
        sys_prompt: str,
        top_p: float = 0.9,
        temperature: float = 0.1,
        max_output_tokens: int = 1024,
        **kwargs
    ):
        self.custom_id = custom_id
        self.user_prompt = user_prompt
        self.sys_prompt = sys_prompt
        self.top_p = top_p
        self.temperature = temperature
        self.max_output_tokens = max_output_tokens
        self.db = db
        self.extra_fields = kwargs  # Store extra fields

    def save(self) -> bool:
        payload = {
            "custom_id": self.custom_id,
            "user_prompt": self.user_prompt,
            "sys_prompt": self.sys_prompt,
            "top_p": self.top_p,
            "temperature": self.temperature,
            "max_output_tokens": self.max_output_tokens,
            "status": "PENDING",
        }
        payload.update(self.extra_fields)  # Merge extra fields into payload
        return self.db.save_payload(**payload)


from .db import Db
from pathlib import Path
from google.cloud import storage
import os
from google.genai.types import CreateBatchJobConfig, HttpOptions
from google import genai


class File:
    def __init__(
        self, db: Db, folder_path: Path, file_name_format: str, gemini_model: str
    ):
        self.folder_path = folder_path
        self.file_name_format = file_name_format
        self.db = db
        self.file_path = None
        self.gemini_model = gemini_model

    def _create(self) -> None:
        try:
            self.folder_path.mkdir(parents=True, exist_ok=True)
            self.file_path = self.folder_path / self.file_name_format
            if not self.file_path.exists():
                self.file_path.touch()
                self.db.save_file(
                    file_path=str(self.file_path.name),
                    file_size=0,
                    status="created"
                )
                print(f"File {self.file_path} created.")
            else:
                print(f"File {self.file_path} already exists.")
        except Exception as e:
            print(f"Error creating file: {e}")

    def _delete(self):
        if self.file_path.exists():
            self.file_path.unlink()
            self.file_path = None
            print(f"File {self.file_path} deleted.")
        else:
            print(f"File {self.file_path} does not exist.")

    def write(self, paylods: list, is_relaunch:bool= False) -> bool:
        try:

            if not paylods:
                return False

            if self.file_path is None:
                self._create()

            if self.file_size_exceed_limits():
                self.process()
                self._create()

            with open(self.file_path, "a") as file:
                for payload in paylods:
                    custom_id = payload['custom_id']

                    if is_relaunch:
                        self.db.update_payload(
                            custom_id= custom_id,
                            relaunched=1
                        )

                    line_content = {
                        "custom_id": payload["custom_id"],
                        "request": {
                            "contents": [
                                {
                                    "role": "user",
                                    "parts": [{"text": payload["user_prompt"]}],
                                }
                            ],
                            "generationConfig": {
                                "temperature": payload["temperature"],
                                "top_p": payload["top_p"],
                                "max_output_tokens": payload["max_output_tokens"],
                            },
                            "systemInstruction": {
                                "parts": [{"text": payload["sys_prompt"]}],
                            },
                        },
                    }
                    file.write(f"{line_content}\n")

                    self.db.update_payload(
                        custom_id=custom_id,
                        status="WRITTEN",
                        file_name=self.file_path.name
                    )

            self.db.update_file(
                file_path=Path(self.file_path.name),
                file_size=self.file_path.stat().st_size
            )

            return True
        except Exception as e:
            print(f"Error writing to file: {e}")
            return False

    def _upload(self, file_path: Path) -> str:
        try:
            # Init client
            client = storage.Client()

            # Get bucket
            bucket_name = os.getenv("GOOGLE_STORAGE_BUCKET")
            bucket = client.bucket(bucket_name)

            # Create blob object inside "input/" folder
            blob = bucket.blob(f"input/gemini/{file_path.name}")

            # Upload file from path
            blob.upload_from_filename(str(file_path))

            return f"gs://{bucket_name}/input/gemini/{file_path.name}"

        except Exception as e:
            print(f"Upload failed: {e}")
            raise

    def process(self):
        try:

            if self.file_path is None:
                print("File not found.")
                return

            google_storage_file = self._upload(self.file_path)

            self.db.update_file(
                file_path=Path(self.file_path.name),
                status="processing"
            )

            vertex_client = genai.Client(
                vertexai=True,
                http_options=HttpOptions(api_version="v1"),
                project=os.getenv("GOOGLE_PROJECT_NAME"),
                location=os.getenv("GOOGLE_PROJECT_LOCATION", "us-central1"),
            )

            output_file = f"gs://{os.getenv('GOOGLE_STORAGE_BUCKET')}/output/gemini/{os.path.splitext(os.path.basename(self.file_path))[0]}"

            vertex_client.batches.create(
                model=self.gemini_model,
                src=google_storage_file,
                config=CreateBatchJobConfig(dest=output_file),
            )

            self._delete()

        except Exception as e:
            print(e)

    def file_size_exceed_limits(self):
        try:
            if self.file_path is None or not self.file_path.exists():
                print("File does not exist.")
                return False
            file_size = self.file_path.stat().st_size
            max_size = int(os.getenv("BATCH_FILE_SIZE_LIMIT")) * 1024 * 1024  # 10 MB in bytes
            if file_size > max_size:
                print(f"File size {file_size} bytes exceeds 10 MB.")
                return True
            else:
                print(f"File size {file_size} bytes is within the 10 MB limit.")
                return False
        except Exception as e:
            print(e)
    
    @staticmethod
    def download(google_storage_file_path: Path, destination_dir: Path) -> Path:
        try:
            client = storage.Client()
            bucket = client.bucket(os.getenv("GOOGLE_STORAGE_BUCKET"))
            blob = bucket.blob(str(google_storage_file_path))

            # Split the GCS path into parts
            blob_parts = str(google_storage_file_path).strip("/").split("/")

            # Get the third part for prefix (index 2)
            folder_name = blob_parts[2] if len(blob_parts) >= 3 else "unknown"

            # Get the original filename
            original_filename = os.path.basename(str(google_storage_file_path))

            # Build the dynamic filename
            dynamic_filename = f"{folder_name}_{original_filename}"

            # Ensure destination directory exists
            os.makedirs(destination_dir, exist_ok=True)

            output_file_path = destination_dir / dynamic_filename

            # Download file
            blob.download_to_filename(output_file_path)

            return output_file_path

        except Exception as e:
            print(e)
            return None



from fastapi import FastAPI, Request
from pathlib import Path
import uvicorn
import json_repair
import asyncio
from .file import File
from .db import Db
import os

class Callback:

    def __init__(self, db:Db, port: int, destination_dir: Path, func: callable = None):
        self.port = port
        self.app = FastAPI()
        self.destination_dir = destination_dir
        self.func = func
        self.db = db

        # register routes
        self.app.post("/batch_processing_done")(self.callback)

    def _process_file_gemini(self, file_path: Path) -> None:
        try:

            if file_path.suffix != ".jsonl" or not file_path.is_relative_to(Path("output/")):
                raise ValueError("File must be a .jsonl file and starts with output/")

            downloaded_file_path = File.download(
                google_storage_file_path=file_path,
                destination_dir=self.destination_dir
            )
            if not downloaded_file_path:
                raise Exception("Failed to download file from Google Cloud Storage")

            with open(downloaded_file_path, "r") as file:
                for line in file:
                    data = json_repair.loads(line)

                    # Skip if status is present (indicating an error)
                    if data.get("status"):
                        self.db.update_payload(
                            custom_id=data.get('custom_id'),
                            status="FAILED"
                        )
                        continue

                    response_brut = data.get("response")["candidates"][0]["content"][
                        "parts"
                    ][0]["text"]
                    response = (
                        response_brut.replace("```json", "").replace("```", "").strip()
                    )

                    self.db.update_payload(
                        custom_id=data.get('custom_id',''),
                        llm_response=json_repair.loads(response),
                        tokens=data.get("response")["usageMetadata"].get("totalTokenCount", 0),
                        status="DONE"
                    )

            os.remove(downloaded_file_path)

        except Exception as e:
            print(e)

    async def callback(self, request: Request):
        try:
            data = await request.json()
            file_path = data.get("name")

            if not file_path:
                raise ValueError("File path is required")

            await asyncio.to_thread(
                self._process_file_gemini, Path(file_path)
            )
            
            # custom func accepts : results as list and file Path
            if self.func:
                await asyncio.to_thread(self.func, Path(file_path))
            
            self.db.update_file(
                file_path=Path(f"{Path(file_path).parts[2]}.jsonl"),
                status="DONE"
            )

            local_file_path = self.destination_dir / Path(file_path).name
            local_file_path.unlink(missing_ok=True)

            return {"status": "success"}

        except Exception as e:
            return {"status": "error", "message": str(e)}

    def start_server(self):
        try:
            uvicorn.run(app=self.app, host="0.0.0.0", port=self.port)
        except Exception as e:
            print(e)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="vertex_batch",
    version="0.1.8",
    author="AYOUB ERRKHIS",
    author_email="ayoub.errkhis@aol.com",
    description="A module for batch processing with Google Cloud Storage and MongoDB integration.",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "fastapi==0.116.1",
        "pymongo==4.13.2",
        "google-cloud-storage==2.19.0",
        "uvicorn==0.35.0",
        "google-genai==1.28.0",
        "json-repair==0.25.3",
        "redis==4.5.2"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.6',
    long_description=long_description,
    long_description_content_type="text/markdown",
)